import sqlite3, contextlib
from dataclasses import dataclass 
from pathlib import Path

_con = None 

def connect(path="data/app.db"):
    global _con 
    # Ensure directory exists
    Path(path).parent.mkdir(parents=True, exist_ok=True)
    _con = sqlite3.connect(path, isolation_level=None, detect_types=sqlite3.PARSE_DECLTYPES)
    _con.row_factory = sqlite3.Row 
    _con.execute("PRAGMA foreign_keys=ON;")
    _con.execute("PRAGMA journal_mode=WAL;")
    _con.execute("PRAGMA busy_timeout=3000;")
    # Auto-apply schema on first connection (idempotent)
    ensure_schema()
    return _con

def ensure_schema():
    """Auto-apply schema.sql on first connection (idempotent)"""
    if _con is None:
        return
    schema_path = Path("data/schema.sql")
    if schema_path.exists():
        try:
            _con.executescript(schema_path.read_text(encoding="utf-8"))
        except sqlite3.Error:
            pass  # Already applied or error - continue 

@dataclass 
class ExecResult:
    rowcount: int
    last_id: int | None 

def _cur():
    if _con is None: connect()
    return _con 

def query(sql, *args):
    cur = _cur().execute(sql, args)
    return [dict(r) for r in cur.fetchall()]

def one(sql, *args):
    rows = query(sql, *args)
    if len(rows) > 1: raise ValueError("one() got %d rows" %len(rows))
    return rows[0] if rows else None 

def exec(sql, *args):
    cur = _cur().execute(sql, args)
    return ExecResult(rowcount=cur.rowcount, last_id=cur.lastrowid)

@contextlib.contextmanager 
def tx():
    try:
        _cur().execute("BEGIN")
        yield
        _cur().execute("COMMIT")
    except Exception:
        _cur().execute("ROLLBACK")
        raise 



from __future__ import annotations
import os
from pathlib import Path
from jinja2 import Environment, FileSystemLoader, select_autoescape

from .app import Response
from .csrf import ensure_token
from .flash import pull

_jinja_env = None
_views_path = None

def _resolve_views_path() -> Path:
    """Prefer ./app/views, then ./views. Allow override via SHIPY_VIEWS."""
    if os.getenv("SHIPY_VIEWS"):
        return Path(os.getenv("SHIPY_VIEWS")).resolve()
    base = Path(os.getenv("SHIPY_BASE", Path.cwd()))
    for candidate in (base / "app" / "views", base / "views"):
        if candidate.exists():
            return candidate.resolve()
    # default fallback even if missing (Jinja will error on missing template)
    return (base / "app" / "views").resolve()

def _env() -> Environment:
    global _jinja_env, _views_path
    if _jinja_env is None:
        _views_path = _resolve_views_path()
        _jinja_env = Environment(
            loader=FileSystemLoader(str(_views_path)),
            autoescape=select_autoescape(["html", "xml"]),
            auto_reload=True,  # fine for dev; Jinja caches in prod anyway
            enable_async=False,
        )
    return _jinja_env

def render(template: str, **ctx) -> Response:
    """Plain render: does NOT touch session/cookies. Good for static pages."""
    html = _env().get_template(template).render(**ctx)
    return Response.html(html)

def render_req(req, template: str, **ctx) -> Response:
    """
    Render with request context:
    - Ensures a CSRF token exists (and sets it in a cookie-backed session)
    - Pulls flash messages (consumes them)
    Exposes `csrf` and `flashes` to the template.
    """
    resp = Response.html("")                 # we need a Response to set cookies
    csrf = ensure_token(req, resp)           # may update session cookie
    flashes = pull(req, resp)                # consumes flash from session
    html = _env().get_template(template).render(csrf=csrf, flashes=flashes, **ctx)
    resp.body = html.encode()
    return resp



import os, base64, hmac
from .session import get_session, set_session 
from .app import Response

def _nonce(n=16): return base64.urlsafe_b64encode(os.urandom(n)).decode().rstrip("=")

def ensure_token(req, resp: Response | None = None) -> str:
    s = get_session(req)
    tok = s.get("csrf")
    if not tok:
        tok = _nonce()
        if resp is None:
            # caller will see it later 
            s["csrf"] = tok 
        else:
            s["csrf"] = tok
            set_session(resp, s)
    return tok 

def verify(req):
    # ensure request.form is loaded before calling verify
    s = get_session(req)
    sent = req.form.get("csrf")
    print("req", req)
    good = bool(sent) and sent == s.get("csrf")
    if not good:
        from .app import Response
        return Response.text("Forbidden (CSRF)", 403)
    return None 

# <input type="hidden" name="csrf" value="{{ csrf }}">



# shipy/app.py
import re, inspect, os, mimetypes, traceback, html
from urllib.parse import parse_qs
from http import cookies as http_cookies
from pathlib import Path

# CSRF needs the session cookie; session.py must NOT import from app.py (no cycles).
from .session import get_session

# ---- Config & paths ---------------------------------------------------------
def _env_bool(name: str, default: bool = False) -> bool:
    v = os.getenv(name)
    if v is None:
        return default
    return v.strip().lower() in ("1", "true", "yes", "on")

DEBUG = _env_bool("SHIPY_DEBUG", True)  # dev default True; set SHIPY_DEBUG=0 in prod

_BASE = Path(os.getenv("SHIPY_BASE", Path.cwd()))
PUBLIC_DIR = Path(os.getenv("SHIPY_PUBLIC", _BASE / "public")).resolve()

# Where to look for error templates
_ERROR_DIRS = [
    (_BASE / "app" / "views" / "errors").resolve(),
    (_BASE / "views" / "errors").resolve(),
]

UNSAFE_METHODS = {"POST", "PUT", "PATCH", "DELETE"}


# ---- Helpers ----------------------------------------------------------------
def _read_error_template(name: str) -> str | None:
    """Return the error HTML if found (e.g., '404.html' in views/errors/)."""
    for d in _ERROR_DIRS:
        f = d / f"{name}.html"
        if f.is_file():
            try:
                return f.read_text(encoding="utf-8")
            except Exception:
                pass
    return None


async def _serve_static(scope, receive, send):
    """Tiny dev-time static server for /public/*."""
    path = scope["path"]
    method = scope["method"].upper()

    if not path.startswith("/public/"):
        return await Response.text("Not Found", 404)(scope, receive, send)

    rel = path[len("/public/"):]
    root = PUBLIC_DIR
    file = (root / rel).resolve()

    # prevent traversal and 404 if missing
    if not str(file).startswith(str(root)) or not file.is_file():
        return await Response.text("Not Found", 404)(scope, receive, send)

    data = file.read_bytes()
    ctype = mimetypes.guess_type(str(file))[0] or "application/octet-stream"
    resp = Response(data, 200, headers=[(b"content-type", ctype.encode())])
    if method == "HEAD":
        resp.body = b""
    return await resp(scope, receive, send)


def _pretty_tb_html(exc: BaseException) -> str:
    tb = "".join(traceback.format_exception(type(exc), exc, exc.__traceback__))
    esc = html.escape(tb)
    return f"""<!doctype html>
<html><head><meta charset="utf-8"><title>500 Internal Server Error</title>
<style>
body{{font:14px/1.45 system-ui,-apple-system,Segoe UI,Roboto,sans-serif;padding:24px;background:#0b0b0b;color:#f6f6f6}}
pre{{background:#111;color:#f6f6f6;padding:16px;border-radius:12px;overflow:auto}}
h1{{margin-top:0}}
a{{color:#60a5fa}}
</style>
</head><body>
<h1>500 Internal Server Error</h1>
<p>DEBUG is on (SHIPY_DEBUG=1). Here’s the traceback:</p>
<pre>{esc}</pre>
</body></html>"""


# ---- Core -------------------------------------------------------------------
class App:
    def __init__(self):
        self.routes = []  # list of (method, compiled_regex, handler)

    def _compile_path(self, path: str) -> re.Pattern:
        # {id:int}  -> (?P<id>\d+)
        def repl_int(m): return f"(?P<{m.group(1)}>\\d+)"
        # {slug}    -> (?P<slug>[^/]+)
        def repl_str(m): return f"(?P<{m.group(1)}>[^/]+)"
        path = re.sub(r"{(\w+):int}", repl_int, path)
        path = re.sub(r"{(\w+)}", repl_str, path)
        return re.compile("^" + path + "$")

    def add(self, method, path, handler):
        self.routes.append((method.upper(), self._compile_path(path), handler))

    def get(self, path, handler):   self.add("GET",  path, handler)
    def post(self, path, handler):  self.add("POST", path, handler)
    # (add put/patch/delete helpers later if needed)

    async def __call__(self, scope, receive, send):
        if scope["type"] != "http":
            return

        method = scope["method"].upper()
        path   = scope["path"]

        # Health check (always 200 ok)
        if path == "/health":
            body = b"ok" if method != "HEAD" else b""
            return await Response(body, 200, headers=[(b"content-type", b"text/plain; charset=utf-8")])(scope, receive, send)

        # Dev static under /public/
        if path.startswith("/public/") and method in ("GET", "HEAD"):
            return await _serve_static(scope, receive, send)

        # Route matching + method handling (405 + HEAD shim)
        allowed: set[str] = set()
        chosen = None          # (handler, params, head_shim: bool)
        for m, rx, handler in self.routes:
            mobj = rx.match(path)
            if not mobj:
                continue
            allowed.add(m)
            if m == method:
                chosen = (handler, mobj.groupdict(), False)
                break
            # HEAD shim: if method is HEAD and route defines GET
            if method == "HEAD" and m == "GET" and chosen is None:
                chosen = (handler, mobj.groupdict(), True)

        if chosen:
            handler, params, head_shim = chosen
            req = Request(scope, receive, params)

            # Global CSRF guard for unsafe methods (form posts)
            if method in UNSAFE_METHODS:
                await req.load_body()
                s = get_session(req) or {}
                sent = req.form.get("csrf")
                if not sent or sent != s.get("csrf"):
                    return await Response.text("Forbidden (CSRF)", 403)(scope, receive, send)

            # Call handler with error handling
            try:
                result = handler(req)
                if inspect.isawaitable(result):
                    result = await result
            except Exception as exc:
                if DEBUG:
                    html_page = _pretty_tb_html(exc)
                    return await Response.html(html_page, 500)(scope, receive, send)
                tpl = _read_error_template("500")
                return await (Response.html(tpl, 500) if tpl else Response.text("Internal Server Error", 500))(scope, receive, send)

            if not isinstance(result, Response):
                result = Response.html(str(result))
            if head_shim:
                result.body = b""
            return await result(scope, receive, send)

        # If we matched the path but not the method → 405 with Allow
        if allowed:
            allow = set(allowed)
            if "GET" in allow:
                allow.add("HEAD")  # advertise HEAD when GET exists
            allow_hdr = ", ".join(sorted(allow))
            return await Response.text("Method Not Allowed", 405)(
                {**scope, "headers": scope.get("headers", [])},  # pass-through
                receive,
                lambda msg: send({**msg, "headers": (msg.get("headers") or []) + [(b"allow", allow_hdr.encode())]})
            )

        # No route matched → 404 page
        tpl = _read_error_template("404")
        return await (Response.html(tpl, 404) if tpl else Response.text("Not Found", 404))(scope, receive, send)


class Response:
    def __init__(self, body=b"", status=200, headers=None, content_type="text/html; charset=utf-8"):
        self.body = body if isinstance(body, bytes) else body.encode()
        self.status = status
        self.headers = headers or [(b"content-type", content_type.encode())]
        self._cookies = http_cookies.SimpleCookie()

    def set_cookie(self, name, value, *, http_only=True, samesite="Lax", path="/", max_age=None, secure=False):
        # In prod, default to Secure cookies unless caller explicitly wants otherwise
        if not DEBUG:
            secure = True
        self._cookies[name] = value
        morsel = self._cookies[name]
        morsel["path"] = path
        morsel["samesite"] = samesite
        if http_only: morsel["httponly"] = True
        if secure: morsel["secure"] = True
        if max_age is not None: morsel["max-age"] = str(max_age)

    def delete_cookie(self, name, path="/"):
        self.set_cookie(name, "", max_age=0, path=path)

    async def __call__(self, scope, receive, send):
        # Default security headers in prod (add if missing)
        if not DEBUG:
            have = {k.lower(): True for (k, _) in ((h[0].decode(), h[1]) for h in self.headers)}
            sec = []
            if "x-content-type-options" not in have:
                sec.append((b"x-content-type-options", b"nosniff"))
            if "referrer-policy" not in have:
                sec.append((b"referrer-policy", b"no-referrer"))
            if "x-frame-options" not in have:
                sec.append((b"x-frame-options", b"DENY"))
            self.headers = list(self.headers) + sec

        headers = list(self.headers)
        for morsel in self._cookies.values():
            headers.append((b"set-cookie", morsel.OutputString().encode()))
        await send({"type": "http.response.start", "status": self.status, "headers": headers})
        await send({"type": "http.response.body", "body": self.body})

    @classmethod
    def html(cls, text, status=200):       return cls(text, status)
    @classmethod
    def text(cls, text, status=200):       return cls(text, status, content_type="text/plain; charset=utf-8")
    @classmethod
    def redirect(cls, location, status=303):
        return cls(b"", status, headers=[(b"location", location.encode())])


class Request:
    def __init__(self, scope, receive, path_params):
        self.scope = scope
        self._receive = receive
        self.method = scope["method"]
        self.path = scope["path"]
        self.query = {
            k: (v[0] if len(v) == 1 else v)
            for k, v in parse_qs(scope.get("query_string", b"").decode()).items()
        }
        self.path_params = path_params
        self._body = None
        self.form = {}
        self.cookies = {}
        for k, v in scope.get("headers", []):
            if k.lower() == b"cookie":
                jar = http_cookies.SimpleCookie()
                jar.load(v.decode())
                self.cookies = {n: morsel.value for n, morsel in jar.items()}
                break

    async def load_body(self):
        if self._body is not None:
            return
        chunks = []
        while True:
            event = await self._receive()
            if event["type"] == "http.request":
                if event.get("body"):
                    chunks.append(event["body"])
                if not event.get("more_body"):
                    break
        self._body = b"".join(chunks)
        headers = {k.decode(): v.decode() for k, v in self.scope.get("headers", [])}
        ctype = headers.get("content-type", "")
        if "application/x-www-form-urlencoded" in ctype:
            self.form = {
                k: (v[0] if len(v) == 1 else v)
                for k, v in parse_qs(self._body.decode()).items()
            }



from .session import get_session, set_session 

def add(req, resp, message, kind="info"):
    s = get_session(req)
    s.setdefault("_flash", []).append({"kind": kind, "msg": message})
    set_session(resp, s)

def pull(req, resp):
    s = get_session(req)
    out = s.get("_flash", [])
    if out:
        s["_flash"] = []
        set_session(resp, s)
    return out 


# shipy/forms.py
from __future__ import annotations
import re

_EMAIL_RE = re.compile(r"^[^@\s]+@[^@\s]+\.[^@\s]+$")

class Form:
    """
    Tiny helper for classic HTML forms.
    - Build from request.form (urlencoded)
    - Chain simple validators
    - Refill fields in templates: {{ form['email'] }}
    - Show errors: {% for e in form.errors_for('email') %}...{% endfor %}
    """
    def __init__(self, data: dict[str, str | list[str]] | None = None):
        data = data or {}
        # normalize multi-values to a single string (first)
        self.data: dict[str, str] = {
            k: (v if isinstance(v, str) else (v[0] if v else ""))
            for k, v in data.items()
        }
        self.errors: dict[str, list[str]] = {}

    # ----- API -----
    @property
    def ok(self) -> bool:
        return not self.errors

    def require(self, *fields: str) -> "Form":
        for f in fields:
            v = self.data.get(f, "").strip()
            if v == "":
                self._err(f, "required")
        return self

    def min(self, field: str, n: int) -> "Form":
        if len(self.data.get(field, "")) < n:
            self._err(field, f"min {n} chars")
        return self

    def email(self, field: str = "email") -> "Form":
        v = self.data.get(field, "").strip()
        if v and not _EMAIL_RE.match(v):
            self._err(field, "invalid email")
        return self

    # templating sugar
    def __getitem__(self, key: str) -> str:
        return self.data.get(key, "")

    def errors_for(self, field: str) -> list[str]:
        return self.errors.get(field, [])

    def to_dict(self) -> dict[str, str]:
        return dict(self.data)

    # ----- internals -----
    def _err(self, field: str, msg: str):
        self.errors.setdefault(field, []).append(msg)



# shipy/session.py
from __future__ import annotations
import json, hmac, hashlib, base64, os

SECRET = os.getenv("SHIPY_SECRET", "dev-secret-change-me").encode()
COOKIE_NAME = "shipy"

def _sign(b: bytes) -> bytes:
    return hmac.new(SECRET, b, hashlib.sha256).digest()

def _pack(data: dict) -> str:
    raw = json.dumps(data, separators=(",", ":")).encode()
    sig = _sign(raw)
    return base64.urlsafe_b64encode(raw + sig).decode()

def _unpack(token: str) -> dict | None:
    if not token:
        return None
    try:
        blob = base64.urlsafe_b64decode(token.encode())
        raw, sig = blob[:-32], blob[-32:]
        if hmac.compare_digest(sig, _sign(raw)):
            return json.loads(raw.decode())
    except Exception:
        pass
    return None

def get_session(req) -> dict:
    """Read & verify the signed cookie from the request; return {} if missing/bad."""
    return _unpack(req.cookies.get(COOKIE_NAME, "")) or {}

def set_session(resp, data: dict):
    """Write session data into a signed cookie on the response."""
    resp.set_cookie(COOKIE_NAME, _pack(data), http_only=True, samesite="Lax")

def clear_session(resp):
    """Clear the session cookie."""
    resp.delete_cookie(COOKIE_NAME)



# shipy/cli.py
from __future__ import annotations

import argparse
import datetime
import os
import secrets
import sqlite3
from pathlib import Path
from textwrap import dedent
from typing import Optional

from shipy.sql import connect as sql_connect  # reuse PRAGMAs/row_factory


# ---------- DB ----------------------------------------------------------------

def cmd_db_init(db_path: str = "data/app.db", schema_path: str = "data/schema.sql") -> int:
    db = Path(db_path)
    db.parent.mkdir(parents=True, exist_ok=True)
    con = sql_connect(str(db))
    schema = Path(schema_path)
    if schema.exists():
        con.executescript(schema.read_text(encoding="utf-8"))
        print(f"✅ DB initialized at {db} (schema: {schema})")
    else:
        print(f"✅ DB created at {db} (no schema.sql found)")
    return 0


def cmd_db_backup(db_path: str = "data/app.db", out_dir: str = "data/backups") -> int:
    srcp = Path(db_path)
    if not srcp.exists():
        print(f"DB not found: {srcp}")
        return 2
    out = Path(out_dir)
    out.mkdir(parents=True, exist_ok=True)
    ts = datetime.datetime.now().strftime("%Y%m%d-%H%M%S")
    dest = out / f"{srcp.stem}-{ts}.sqlite"
    src = sqlite3.connect(str(srcp))
    dst = sqlite3.connect(str(dest))
    with dst:
        src.backup(dst)
    dst.close()
    src.close()
    print(f"📦 Backup written: {dest}")
    return 0


# ---------- DEV ----------------------------------------------------------------

def cmd_dev(app_ref: str, host: str, port: int, reload: bool, workers: int, show_info: bool = True) -> int:
    # Ensure the current working directory is importable as top-level (so 'app.*' resolves)
    import sys
    import uvicorn  # type: ignore

    cwd = os.getcwd()
    os.environ.setdefault("SHIPY_BASE", cwd)

    if cwd not in sys.path:
        sys.path.insert(0, cwd)

    # Ensure the reload subprocess inherits the path too
    os.environ["PYTHONPATH"] = cwd + os.pathsep + os.environ.get("PYTHONPATH", "")

    if show_info:
        print(f"🚀 Shipy dev server starting...")
        print(f"📁 Working directory: {cwd}")
        print(f"🗄️  Database: data/app.db")
        print(f"📋 Schema: data/schema.sql")
        print(f"🌐 URL: http://{host}:{port}")
        print("Press Ctrl+C to stop\n")

    config = uvicorn.Config(
        app_ref,
        host=host,
        port=port,
        reload=reload,
        workers=workers,
        reload_dirs=[cwd],
    )
    server = uvicorn.Server(config)
    return 0 if server.run() else 1


# ---------- NEW ----------------------------------------------------------------

def _write_file(path: Path, content: str, *, force: bool):
    if path.exists() and not force:
        print(f"• skip {path} (exists)")
        return
    path.parent.mkdir(parents=True, exist_ok=True)
    path.write_text(dedent(content).lstrip(), encoding="utf-8")
    print(f"✓ write {path}")


def cmd_new(name: str, *, force: bool = False) -> int:
    root = Path(name).resolve()
    app = root / "app"
    views = app / "views"
    public = root / "public"
    dbdir = root / "data"

    files: dict[Path, str] = {
        root / ".gitignore": """
            __pycache__/
            *.pyc
            .env
            data/*.db*
            .DS_Store
        """,
        # Make 'app' a real package so 'app.main:app' imports everywhere
        app / "__init__.py": "",
        public / "base.css": """
            :root { --bg:#0b0b0b; --fg:#f6f6f6; --muted:#b7b7b7; --card:#151515; --acc:#60a5fa; }
            *{box-sizing:border-box} body{margin:0;background:var(--bg);color:var(--fg);font:16px/1.45 system-ui,-apple-system,Segoe UI,Roboto,sans-serif}
            .wrap{max-width:720px;margin:40px auto;padding:0 16px}
            a{color:var(--acc);text-decoration:none}
            header{display:flex;justify-content:space-between;align-items:center;margin-bottom:20px}
            .card{background:var(--card);padding:16px;border-radius:12px;margin:12px 0}
            .stack>*+*{margin-top:10px}
            .input, textarea{width:100%;padding:10px;border-radius:10px;border:1px solid #333;background:#111;color:var(--fg)}
            .btn{display:inline-block;border:0;border-radius:10px;padding:10px 14px;background:#2563eb;color:#fff;cursor:pointer}
            .err{color:#f87171;font-size:14px}
            .flash{padding:8px 12px;border-radius:10px;background:#0f172a;color:#e2e8f0}
            nav a{margin-right:10px}
        """,
        app / "main.py": """
            from shipy.app import App, Response
            from shipy.render import render_req
            from shipy.sql import connect, query, one, exec, tx
            from shipy.forms import Form
            from shipy.auth import (
                current_user, require_login,
                hash_password, check_password,
                login, logout,
                too_many_login_attempts, record_login_failure, reset_login_failures
            )

            app = App()
            connect("data/app.db")

            def home(req):
                user = current_user(req)
                return render_req(req, "home/index.html", user=user)

            def signup_form(req): return render_req(req, "users/new.html")

            async def signup(req):
                await req.load_body()
                form = Form(req.form).require("email","password").min("password", 6).email("email")
                if not form.ok: return render_req(req, "users/new.html", form=form)
                if one("SELECT id FROM users WHERE email=?", form["email"]):
                    form.errors.setdefault("email", []).append("already registered")
                    return render_req(req, "users/new.html", form=form)
                with tx():
                    exec("INSERT INTO users(email,password_hash) VALUES(?,?)",
                         form["email"], hash_password(form['password']))
                    u = one("SELECT id,email FROM users WHERE email=?", form["email"])
                resp = Response.redirect("/")
                login(req, resp, u["id"])
                return resp

            def login_form(req): return render_req(req, "sessions/login.html")

            async def login_post(req):
                await req.load_body()
                form = Form(req.form).require("email","password").email("email")
                ip = req.scope.get("client", ("",0))[0] or "unknown"
                if too_many_login_attempts(ip):
                    form.errors.setdefault("email", []).append("too many attempts, try later")
                    return render_req(req, "sessions/login.html", form=form)
                u = one("SELECT id,email,password_hash FROM users WHERE email=?", form["email"])
                if not u or not check_password(form["password"], u["password_hash"]):
                    record_login_failure(ip)
                    form.errors.setdefault("email", []).append("invalid email or password")
                    return render_req(req, "sessions/login.html", form=form)
                reset_login_failures(ip)
                resp = Response.redirect("/")
                login(req, resp, u["id"])
                return resp

            async def logout_post(req):
                resp = Response.redirect("/")
                logout(resp)
                return resp


            def secret(req):
                user = require_login(req)
                if not user: return Response.redirect("/login")
                return render_req(req, "secret.html", user=user)

            app.get("/", home)
            app.get("/signup", signup_form)
            app.post("/signup", signup)
            app.get("/login", login_form)
            app.post("/login", login_post)
            app.post("/logout", logout_post)
            app.get("/secret", secret)
        """,
        views / "home" / "index.html": """
            <!doctype html><meta charset="utf-8"><link rel="stylesheet" href="/public/base.css">
            <div class="wrap">
              <header>
                <strong>Shipy</strong>
                <nav>
                  {% if user %}
                    <span>{{ user.email }}</span>
                    <form method="post" action="/logout" style="display:inline">
                      <input type="hidden" name="csrf" value="{{ csrf }}"><button class="btn">Logout</button>
                    </form>
                  {% else %}
                    <a href="/login">Login</a> <a href="/signup">Sign up</a>
                  {% endif %}
                </nav>
              </header>

              {% if flashes %}{% for f in flashes %}<div class="flash">{{ f.msg }}</div>{% endfor %}{% endif %}

              <div class="card">
                <h2>Posts</h2>
                <div class="stack">
                  {% for p in posts %}
                    <div class="card"><strong>#{{ p.id }}</strong> — {{ p.title }}</div>
                  {% else %}
                    <div class="muted">No posts yet.</div>
                  {% endfor %}
                </div>
              </div>

              {% if user %}
              <div class="card">
                <h3>New post</h3>
                <form method="post" action="/posts" class="stack">
                  <input class="input" name="title" placeholder="Title" value="{{ form['title'] if form else '' }}">
                  {% if form %}{% for e in form.errors_for('title') %}<div class="err">title: {{ e }}</div>{% endfor %}{% endif %}
                  <textarea class="input" name="body" placeholder="Body">{{ form['body'] if form else '' }}</textarea>
                  {% if form %}{% for e in form.errors_for('body') %}<div class="err">body: {{ e }}</div>{% endfor %}{% endif %}
                  <input type="hidden" name="csrf" value="{{ csrf }}">
                  <button class="btn">Create</button>
                </form>
              </div>
              {% else %}
                <div class="card">Log in to create posts.</div>
              {% endif %}
            </div>
        """,
        views / "sessions" / "login.html": """
            <!doctype html><meta charset="utf-8"><link rel="stylesheet" href="/public/base.css">
            <div class="wrap">
              <h1>Log in</h1>
              <form method="post" action="/login" class="stack">
                <label>Email <input class="input" name="email" value="{{ form['email'] if form else '' }}"></label>
                {% if form %}{% for e in form.errors_for('email') %}<div class="err">email: {{ e }}</div>{% endfor %}{% endif %}
                <label>Password <input class="input" type="password" name="password"></label>
                {% if form %}{% for e in form.errors_for('password') %}<div class="err">password: {{ e }}</div>{% endfor %}{% endif %}
                <input type="hidden" name="csrf" value="{{ csrf }}">
                <button class="btn">Log in</button>
              </form>
              <p>No account? <a href="/signup">Sign up</a></p>
            </div>
        """,
        views / "users" / "new.html": """
            <!doctype html><meta charset="utf-8"><link rel="stylesheet" href="/public/base.css">
            <div class="wrap">
              <h1>Sign up</h1>
              <form method="post" action="/signup" class="stack">
                <label>Email <input class="input" name="email" value="{{ form['email'] if form else '' }}"></label>
                {% if form %}{% for e in form.errors_for('email') %}<div class="err">email: {{ e }}</div>{% endfor %}{% endif %}
                <label>Password (min 6) <input class="input" type="password" name="password"></label>
                {% if form %}{% for e in form.errors_for('password') %}<div class="err">password: {{ e }}</div>{% endfor %}{% endif %}
                <input type="hidden" name="csrf" value="{{ csrf }}">
                <button class="btn">Create account</button>
              </form>
              <p>Have an account? <a href="/login">Log in</a></p>
            </div>
        """,
        views / "secret.html": """
            <!doctype html><meta charset="utf-8"><link rel="stylesheet" href="/public/base.css">
            <div class="wrap"><h1>Secret</h1><p>Hello {{ user.email }}!</p></div>
        """,
        views / "errors" / "404.html": """
            <!doctype html><meta charset="utf-8"><link rel="stylesheet" href="/public/base.css">
            <div class="wrap"><h1>404</h1><p>Page not found.</p></div>
        """,
        views / "errors" / "500.html": """
            <!doctype html><meta charset="utf-8"><link rel="stylesheet" href="/public/base.css">
            <div class="wrap"><h1>Something went wrong</h1><p>Please try again.</p></div>
        """,
        dbdir / "schema.sql": """
            CREATE TABLE IF NOT EXISTS users (
              id INTEGER PRIMARY KEY,
              email TEXT NOT NULL UNIQUE,
              password_hash TEXT NOT NULL,
              created_at TIMESTAMP NOT NULL DEFAULT CURRENT_TIMESTAMP
            );
            CREATE INDEX IF NOT EXISTS idx_users_email ON users(email);
        """,
    }
    for path, content in files.items():
        _write_file(path, content, force=force)

    print("\nNext steps:")
    print(f"  cd {root}")
    print("  shipy db init")
    print("  shipy dev --app app.main:app")
    return 0


# ---------- DEPLOY ------------------------------------------------------------

def cmd_deploy_emit(path: str, service: str, domain: str, port: int, user: str, workdir: str) -> int:
    out = Path(path)
    out.mkdir(parents=True, exist_ok=True)
    workdir = str(Path(workdir).resolve())
    public = str((Path(workdir) / "public").resolve())

    service_file = out / f"{service}.service"
    nginx_file = out / f"{domain}.conf"

    _write_file(service_file, f"""
        [Unit]
        Description=Shipy app {service}
        After=network.target

        [Service]
        Type=simple
        WorkingDirectory={workdir}
        Environment=SHIPY_BASE={workdir}
        Environment=SHIPY_DEBUG=0
        # Replace this with: shipy gensecret
        Environment=SHIPY_SECRET=CHANGE_ME
        ExecStart=/usr/bin/env uvicorn app.main:app --host 127.0.0.1 --port {port} --workers 2
        Restart=always
        User={user}
        Group={user}

        [Install]
        WantedBy=multi-user.target
    """, force=True)

    _write_file(nginx_file, f"""
        server {{
          listen 80;
          server_name {domain};

          root {public};

          location /public/ {{
            try_files $uri =404;
            add_header Cache-Control "public, max-age=31536000, immutable";
          }}

          location / {{
            proxy_pass http://127.0.0.1:{port};
            proxy_http_version 1.1;
            proxy_set_header Host $host;
            proxy_set_header X-Forwarded-For $proxy_add_x_forwarded_for;
            proxy_set_header X-Forwarded-Proto $scheme;
            proxy_redirect off;
          }}
        }}
    """, force=True)

    print(f"\nDeploy files written to: {out}")
    print(f"  • systemd: {service_file.name}")
    print(f"  • nginx  : {nginx_file.name}")
    print("\nServer steps (Ubuntu):")
    print("  sudo cp deploy/*.service /etc/systemd/system/")
    print(f"  sudo systemctl enable --now {service}.service")
    print(f"  sudo cp deploy/{domain}.conf /etc/nginx/sites-available/")
    print(f"  sudo ln -sf /etc/nginx/sites-available/{domain}.conf /etc/nginx/sites-enabled/{domain}.conf")
    print("  sudo nginx -t && sudo systemctl reload nginx")
    print("  curl -s http://YOUR_DOMAIN/health")
    return 0


def cmd_gensecret() -> int:
    print(secrets.token_urlsafe(32))
    return 0


def cmd_version() -> int:
    """Print the current Shipy version."""
    import importlib.metadata
    try:
        version = importlib.metadata.version("shipy-web")
        print(f"shipy-web {version}")
    except importlib.metadata.PackageNotFoundError:
        print("shipy-web (development version)")
    return 0


# ---------- MAIN --------------------------------------------------------------

def main(argv: Optional[list[str]] = None) -> int:
    p = argparse.ArgumentParser(prog="shipy", description="Shipy CLI")
    
    # Add version argument
    p.add_argument("--version", action="store_true", help="Show version and exit")
    
    sub = p.add_subparsers(dest="cmd", required=False)

    p_new = sub.add_parser("new", help="Create a new Shipy app skeleton")
    p_new.add_argument("name", help="Directory name (or path)")
    p_new.add_argument("--force", action="store_true", help="Overwrite existing files")

    p_dev = sub.add_parser("dev", help="Run dev server")
    p_dev.add_argument("--app", default="app.main:app", help="ASGI app path, e.g. app.main:app")
    p_dev.add_argument("--host", default="127.0.0.1")
    p_dev.add_argument("--port", type=int, default=8000)
    p_dev.add_argument("--reload", action="store_true", default=True)
    p_dev.add_argument("--no-reload", dest="reload", action="store_false")
    p_dev.add_argument("--workers", type=int, default=1)
    p_dev.add_argument("--show-info", action="store_true", default=True, help="Show boot info")

    p_db = sub.add_parser("db", help="Database utilities")
    p_db_sub = p_db.add_subparsers(dest="db_cmd", required=True)
    p_db_init = p_db_sub.add_parser("init", help="Create SQLite and run schema.sql (if present)")
    p_db_init.add_argument("--db", default="data/app.db")
    p_db_init.add_argument("--schema", default="data/schema.sql")
    p_db_backup = p_db_sub.add_parser("backup", help="Online backup to db/backups/")
    p_db_backup.add_argument("--db", default="data/app.db")
    p_db_backup.add_argument("--out", default="data/backups")

    p_dep = sub.add_parser("deploy", help="Deployment helpers")
    p_dep_sub = p_dep.add_subparsers(dest="dep_cmd", required=True)
    p_emit = p_dep_sub.add_parser("emit", help="Write systemd + nginx files to ./deploy")
    p_emit.add_argument("--path", default="deploy")
    p_emit.add_argument("--service", default=Path(os.getcwd()).name + "-app")
    p_emit.add_argument("--domain", required=True, help="your-domain.com")
    p_emit.add_argument("--port", type=int, default=8000)
    p_emit.add_argument("--user", default="www-data")
    p_emit.add_argument("--workdir", default=os.getcwd())

    p_secret = sub.add_parser("gensecret", help="Generate a random SHIPY_SECRET")

    args = p.parse_args(argv)

    # Handle version command
    if args.version:
        return cmd_version()

    if args.cmd == "new":
        return cmd_new(args.name, force=args.force)
    if args.cmd == "dev":
        return cmd_dev(args.app, args.host, args.port, args.reload, args.workers, args.show_info)
    if args.cmd == "db" and args.db_cmd == "init":
        return cmd_db_init(args.db, args.schema)
    if args.cmd == "db" and args.db_cmd == "backup":
        return cmd_db_backup(args.db, args.out)
    if args.cmd == "deploy" and args.dep_cmd == "emit":
        return cmd_deploy_emit(args.path, args.service, args.domain, args.port, args.user, args.workdir)
    if args.cmd == "gensecret":
        return cmd_gensecret()

    p.print_help()
    return 0


if __name__ == "__main__":
    raise SystemExit(main())



# shipy/auth.py
from __future__ import annotations
import os, time, secrets, hashlib, hmac, binascii
from typing import Optional

from .session import get_session, set_session, clear_session
from .sql import one, exec as sql_exec, tx

# --- Password hashing ---------------------------------------------------------
# Optional bcrypt; falls back to PBKDF2-HMAC if not installed.
try:
    import bcrypt  # type: ignore
    _HAS_BCRYPT = True
except Exception:
    _HAS_BCRYPT = False

def hash_password(password: str) -> str:
    pw = password.encode()
    if _HAS_BCRYPT:
        return "bcrypt$" + bcrypt.hashpw(pw, bcrypt.gensalt()).decode()
    # PBKDF2-HMAC-SHA256 with 200k iters
    salt = os.urandom(16)
    dk = hashlib.pbkdf2_hmac("sha256", pw, salt, 200_000)
    return "pbkdf2$200000$" + binascii.hexlify(salt).decode() + "$" + binascii.hexlify(dk).decode()

def check_password(password: str, stored: str) -> bool:
    if stored.startswith("bcrypt$"):
        if not _HAS_BCRYPT: return False
        return bcrypt.checkpw(password.encode(), stored[len("bcrypt$"):].encode())
    if stored.startswith("pbkdf2$"):
        _, iters_s, salt_hex, hash_hex = stored.split("$", 3)
        iters = int(iters_s)
        salt = binascii.unhexlify(salt_hex)
        want = binascii.unhexlify(hash_hex)
        got = hashlib.pbkdf2_hmac("sha256", password.encode(), salt, iters)
        return hmac.compare_digest(got, want)
    return False

# --- Current user / session helpers -------------------------------------------
def current_user(req) -> Optional[dict]:
    s = get_session(req)
    if not s or "uid" not in s: return None
    return one("SELECT id, email, created_at FROM users WHERE id=?", s["uid"])

def login(req, resp, user_id: int):
    s = get_session(req) or {}
    s["uid"] = int(user_id)
    s["sv"] = 1  # session version; bump to force logout all
    set_session(resp, s)

def logout(resp):
    clear_session(resp)

def require_login(req):
    """Return user dict or None. Handlers can branch or redirect."""
    return current_user(req)

# --- Simple rate limit for login attempts -------------------------------------
# 5 failures per 5 minutes per IP.
def too_many_login_attempts(ip: str, window_sec=300, limit=5) -> bool:
    now = int(time.time())
    with tx():
        sql_exec("""CREATE TABLE IF NOT EXISTS login_attempts(
            ip TEXT, ts INTEGER
        )""")
        sql_exec("DELETE FROM login_attempts WHERE ts < ?", now - window_sec)
        row = one("SELECT COUNT(*) AS c FROM login_attempts WHERE ip=? AND ts>=?", ip, now - window_sec)
    return (row["c"] if row else 0) >= limit

def record_login_failure(ip: str):
    now = int(time.time())
    with tx():
        sql_exec("INSERT INTO login_attempts(ip, ts) VALUES(?,?)", ip, now)

def reset_login_failures(ip: str):
    with tx():
        sql_exec("DELETE FROM login_attempts WHERE ip=?", ip)



import os, sys, datetime

DEBUG = os.getenv("SHIPY_DEBUG") == "1"

def log(*args):
    """Print only when SHIPY_DEBUG=1 is set"""
    if not DEBUG:
        return
    ts = datetime.datetime.now().strftime("%H:%M:%S")
    print(f"[shipy {ts}]", *args, file=sys.stderr, flush=True)
